@@ -6,6 +6,7 @@ import jsonschema
 import python_jsonschema_objects as pjs
 
 import logging
+
 logging.basicConfig(level=logging.DEBUG)
 
 
@@ -17,10 +18,10 @@ def test_schema_validation():
         "$id": "test",
         "type": "object",
         "properties": {
-                "name": "string",
-                "email": {"oneOf": [{"type": "string"}, {"type": "integer"}]},
+            "name": "string",
+            "email": {"oneOf": [{"type": "string"}, {"type": "integer"}]},
         },
-        "required": ["email"]
+        "required": ["email"],
     }
     with pytest.raises(jsonschema.ValidationError):
         pjs.ObjectBuilder(schema)
@@ -32,10 +33,10 @@ def test_regression_9():
         "$id": "test",
         "type": "object",
         "properties": {
-                "name": {"type": "string"},
-                "email": {"oneOf": [{"type": "string"}, {"type": "integer"}]},
+            "name": {"type": "string"},
+            "email": {"oneOf": [{"type": "string"}, {"type": "integer"}]},
         },
-        "required": ["email"]
+        "required": ["email"],
     }
     builder = pjs.ObjectBuilder(schema)
     builder.build_classes()
@@ -52,16 +53,9 @@ def test_build_classes_is_idempotent():
         },
         "required": ["email"],
         "definitions": {
-            "reffed": {
-                "type": "string"
-                },
-            "foo": {
-                "type": "array",
-                "items": {
-                    "$ref": "#/definitions/reffed"
-                }
-            }
-        }
+            "reffed": {"type": "string"},
+            "foo": {"type": "array", "items": {"$ref": "#/definitions/reffed"}},
+        },
     }
     builder = pjs.ObjectBuilder(schema)
     x = builder.build_classes()
@@ -73,12 +67,7 @@ def test_underscore_properties():
         "$schema": "http://json-schema.org/schema#",
         "title": "AggregateQuery",
         "type": "object",
-        "properties": {
-            "group": {
-                "type": "object",
-                "properties": {}
-            }
-        }
+        "properties": {"group": {"type": "object", "properties": {}}},
     }
 
     builder = pjs.ObjectBuilder(schema)
@@ -86,16 +75,12 @@ def test_underscore_properties():
     my_obj_type = ns.Aggregatequery
     request_object = my_obj_type(
         group={
-            "_id": {"foo_id": "$foo_id",
-                    "foo_type": "$foo_type"},
-            "foo": {"$sum": 1}, }
+            "_id": {"foo_id": "$foo_id", "foo_type": "$foo_type"},
+            "foo": {"$sum": 1},
+        }
     )
 
-    assert request_object.group._id == {
-        "foo_id": "$foo_id",
-        "foo_type": "$foo_type"
-    }
-
+    assert request_object.group._id == {"foo_id": "$foo_id", "foo_type": "$foo_type"}
 
 
 def test_array_regressions():
@@ -104,62 +89,54 @@ def test_array_regressions():
         "$id": "test",
         "type": "object",
         "properties": {
-                "name": {"type": "string"},
-                "email_aliases": {
-                    "type": "object",
-                    "additionalProperties": {
-                        "type": "array",
-                        "items": {"$ref": "#/definitions/foo"}
-                    }
-                }
+            "name": {"type": "string"},
+            "email_aliases": {
+                "type": "object",
+                "additionalProperties": {
+                    "type": "array",
+                    "items": {"$ref": "#/definitions/foo"},
+                },
+            },
         },
-        "definitions": {
-            "foo": {
-                "type": "string"
-            }
-        }
+        "definitions": {"foo": {"type": "string"}},
     }
     builder = pjs.ObjectBuilder(schema)
 
     ns = builder.build_classes()
 
-    x = ns.Test.from_json('''{"email_aliases": {
+    x = ns.Test.from_json(
+        """{"email_aliases": {
             "Freddie": ["james", "bond"]
-            }}''')
+            }}"""
+    )
     x.validate()
 
-    y = ns.Test(email_aliases={
-        "Freddie": ["james", "bond"]
-    })
+    y = ns.Test(email_aliases={"Freddie": ["james", "bond"]})
     y.validate()
 
 
 def test_arrays_can_have_reffed_items_of_mixed_type():
     schema = {
-            "$schema": "http://json-schema.org/schema#",
-            "$id": "test",
-            "type": "object",
-            "properties": {
-                "list": {
-                    "type": "array",
-                    "items": {"oneOf": [
+        "$schema": "http://json-schema.org/schema#",
+        "$id": "test",
+        "type": "object",
+        "properties": {
+            "list": {
+                "type": "array",
+                "items": {
+                    "oneOf": [
                         {"$ref": "#/definitions/foo"},
                         {
                             "type": "object",
-                            "properties": {
-                                "bar": {"type":"string"}
-                                },
-                            "required": ["bar"]
-                        }
-                        ]},
-                    }
+                            "properties": {"bar": {"type": "string"}},
+                            "required": ["bar"],
+                        },
+                    ]
                 },
-            "definitions": {
-                "foo": {
-                    "type": "string"
-                    }
-                }
             }
+        },
+        "definitions": {"foo": {"type": "string"}},
+    }
     builder = pjs.ObjectBuilder(schema)
     ns = builder.build_classes()
 
@@ -173,38 +150,36 @@ def test_regression_39():
     builder = pjs.ObjectBuilder("test/thing-two.json")
     ns = builder.build_classes()
 
-    for thing in ('BarMessage', 'BarGroup', "Bar", "Header"):
+    for thing in ("BarMessage", "BarGroup", "Bar", "Header"):
         assert thing in ns
 
-    x = ns.BarMessage(id="message_id",
-                      title="my bar group",
-                      bars=[{"name": "Freddies Half Shell"}]
-                      )
+    x = ns.BarMessage(
+        id="message_id", title="my bar group", bars=[{"name": "Freddies Half Shell"}]
+    )
 
     x.validate()
 
-    # Now an invalid one 
+    # Now an invalid one
     with pytest.raises(pjs.ValidationError):
-        ns.BarMessage(id="message_id",
-                      title="my bar group",
-                      bars=[{"Address": "I should have a name"}]
-                      )
+        ns.BarMessage(
+            id="message_id",
+            title="my bar group",
+            bars=[{"Address": "I should have a name"}],
+        )
 
 
 def test_loads_markdown_schema_extraction(markdown_examples):
-    assert 'Other' in markdown_examples
+    assert "Other" in markdown_examples
 
 
 def test_object_builder_loads_memory_references(markdown_examples):
-    builder = pjs.ObjectBuilder(
-        markdown_examples['Other'],
-        resolved=markdown_examples)
+    builder = pjs.ObjectBuilder(markdown_examples["Other"], resolved=markdown_examples)
     assert builder
 
     with pytest.raises(pjs.ValidationError):
         builder.validate({"MyAddress": 1234})
 
-    builder.validate({"MyAddress": '1234'})
+    builder.validate({"MyAddress": "1234"})
 
 
 def test_object_builder_reads_all_definitions(markdown_examples):
@@ -215,14 +190,13 @@ def test_object_builder_reads_all_definitions(markdown_examples):
 
 @pytest.fixture
 def oneOf(markdown_examples):
-    builder = pjs.ObjectBuilder(markdown_examples['OneOf'], resolved=markdown_examples)
-    return builder.classes['Oneof']
+    builder = pjs.ObjectBuilder(markdown_examples["OneOf"], resolved=markdown_examples)
+    return builder.classes["Oneof"]
 
 
-@pytest.mark.parametrize('json_object', [
-    '{"MyData": "an address"}',
-    '{"MyData": "1234"}'
-])
+@pytest.mark.parametrize(
+    "json_object", ['{"MyData": "an address"}', '{"MyData": "1234"}']
+)
 def test_oneOf_validates_against_any_valid(oneOf, json_object):
 
     oneOf.from_json(json_object)
@@ -235,14 +209,16 @@ def test_oneOf_fails_against_non_matching(oneOf):
 
 @pytest.fixture
 def oneOfBare(markdown_examples):
-    builder = pjs.ObjectBuilder(markdown_examples['OneOfBare'], resolved=markdown_examples)
-    return builder.classes['Oneofbare']
+    builder = pjs.ObjectBuilder(
+        markdown_examples["OneOfBare"], resolved=markdown_examples
+    )
+    return builder.classes["Oneofbare"]
 
 
-@pytest.mark.parametrize('json_object', [
-    '{"MyAddress": "an address"}',
-    '{"firstName": "John", "lastName": "Winnebago"}'
-])
+@pytest.mark.parametrize(
+    "json_object",
+    ['{"MyAddress": "an address"}', '{"firstName": "John", "lastName": "Winnebago"}'],
+)
 def test_oneOfBare_validates_against_any_valid(oneOfBare, json_object):
 
     oneOfBare.from_json(json_object)
@@ -255,9 +231,9 @@ def test_oneOfBare_fails_against_non_matching(oneOfBare):
 
 @pytest.fixture
 def Other(markdown_examples):
-    builder = pjs.ObjectBuilder(markdown_examples['Other'], resolved=markdown_examples)
+    builder = pjs.ObjectBuilder(markdown_examples["Other"], resolved=markdown_examples)
     assert builder
-    return builder.classes['Other']
+    return builder.classes["Other"]
 
 
 def test_additional_props_allowed_by_default(Person):
@@ -267,10 +243,12 @@ def test_additional_props_allowed_by_default(Person):
 
 
 def test_additional_props_permitted_explicitly(markdown_examples):
-    builder = pjs.ObjectBuilder(markdown_examples['AddlPropsAllowed'], resolved=markdown_examples)
+    builder = pjs.ObjectBuilder(
+        markdown_examples["AddlPropsAllowed"], resolved=markdown_examples
+    )
     assert builder
 
-    test = builder.classes['Addlpropsallowed']()
+    test = builder.classes["Addlpropsallowed"]()
     test.randomAttribute = 40
     assert int(test.randomAttribute) == 40
 
@@ -278,16 +256,27 @@ def test_additional_props_permitted_explicitly(markdown_examples):
 def test_still_raises_when_accessing_undefined_attrs(Person):
 
     person = Person()
+    person.firstName = "James"
+
+    # If the attribute doesn't exist, we expect an AttributeError
     with pytest.raises(AttributeError):
         print(person.randomFoo)
 
+    # If the attribute is literal-esq, isn't set but isn't required, accessing it should be fine
+    assert person.gender == None
+
+    # If the attribute is an object, isn't set, but isn't required accessing it should throw an exception
+    with pytest.raises(AttributeError) as e:
+        print(person.address.street)
+        assert "'NoneType' object has no attribute 'street'" in e
+
 
 def test_permits_deletion_of_additional_properties(Person):
     person = Person()
     person.randomthing = 44
     assert int(person.randomthing) == 44
 
-    del person['randomthing']
+    del person["randomthing"]
 
     with pytest.raises(AttributeError):
         assert person.randomthing is None
@@ -315,12 +304,11 @@ def test_object_equality_should_compare_data(Person):
 
 
 def test_object_allows_attributes_in_oncstructor(Person):
-    person = Person(firstName="James",
-            lastName="Bond", age=35)
+    person = Person(firstName="James", lastName="Bond", age=35)
 
     assert person
-    assert str(person.firstName) == 'James'
-    assert str(person.lastName) == 'Bond'
+    assert str(person.firstName) == "James"
+    assert str(person.lastName) == "Bond"
     assert int(person.age) == 35
 
 
@@ -357,12 +345,11 @@ def test_validation_of_mixed_type_enums(Person):
 
 
 def test_objects_allow_non_required_attrs_to_be_missing(Person):
-    person = Person(firstName="James",
-                    lastName="Bond")
+    person = Person(firstName="James", lastName="Bond")
 
     assert person
-    assert str(person.firstName) == 'James'
-    assert str(person.lastName) == 'Bond'
+    assert str(person.firstName) == "James"
+    assert str(person.lastName) == "Bond"
 
 
 def test_objects_require_required_attrs_on_validate(Person):
@@ -374,21 +361,21 @@ def test_objects_require_required_attrs_on_validate(Person):
         person.validate()
 
 
-@pytest.fixture(scope='function')
+@pytest.fixture(scope="function")
 def person_object(Person):
-    return Person(firstName='James')
+    return Person(firstName="James")
 
 
 def test_attribute_access_via_dict(person_object):
-    name = person_object['firstName']
-    assert str(name) == 'James'
+    name = person_object["firstName"]
+    assert str(name) == "James"
 
 
 def test_attribute_set_via_dict(person_object):
-    person_object['firstName'] = "John"
+    person_object["firstName"] = "John"
 
-    name = person_object['firstName']
-    assert str(name) == 'John'
+    name = person_object["firstName"]
+    assert str(name) == "John"
 
 
 def test_numeric_attribute_validation(Person):
@@ -403,7 +390,7 @@ def test_numeric_attribute_validation(Person):
 def test_objects_validate_prior_to_serializing(Person):
     p = Person(firstName="James", lastName="Bond")
 
-    p._properties['age'] = -1
+    p._properties["age"] = -1
 
     with pytest.raises(pjs.ValidationError):
         p.serialize()
@@ -413,61 +400,54 @@ def test_serializing_removes_null_objects(Person):
     person = Person(firstName="James", lastName="Bond")
 
     json_str = person.serialize()
-    assert 'age' not in json.loads(json_str)
+    assert "age" not in json.loads(json_str)
 
 
 def test_lists_get_serialized_correctly(Person):
-    person = Person(
-        firstName="James",
-        lastName="Bond",
-        dogs=["Lassie", "Bobo"]
-    )
+    person = Person(firstName="James", lastName="Bond", dogs=["Lassie", "Bobo"])
 
     json_str = person.serialize()
     assert json.loads(json_str) == {
-        'firstName': "James",
-        'lastName': "Bond",
-        'dogs': ["Lassie", "Bobo"]
+        "firstName": "James",
+        "lastName": "Bond",
+        "dogs": ["Lassie", "Bobo"],
     }
 
 
-@pytest.mark.parametrize('pdict', [
-    dict(
-        firstName="James",
-        lastName="Bond",
-        dogs=["Lassie", "Bobo"]
-    ),
-    dict(
-        firstName="James",
-        lastName="Bond",
-        address={
-            "street": "10 Example Street",
-            "city": "Springfield",
-            "state": "USA"
-        },
-        dogs=["Lassie", "Bobo"]
-    )
-])
+@pytest.mark.parametrize(
+    "pdict",
+    [
+        dict(firstName="James", lastName="Bond", dogs=["Lassie", "Bobo"]),
+        dict(
+            firstName="James",
+            lastName="Bond",
+            address={
+                "street": "10 Example Street",
+                "city": "Springfield",
+                "state": "USA",
+            },
+            dogs=["Lassie", "Bobo"],
+        ),
+    ],
+)
 def test_dictionary_transformation(Person, pdict):
 
     person = Person(**pdict)
 
     assert person.as_dict() == pdict
 
+
 def test_strict_mode():
     schema = {
         "$schema": "http://json-schema.org/schema#",
         "type": "object",
-        "properties": {
-                "firstName": {"type": "string"},
-                "lastName": {"type":"string"},
-        },
+        "properties": {"firstName": {"type": "string"}, "lastName": {"type": "string"}},
         "$id": "test",
-        "title":"Name Data",
-        "required": ["firstName"]
+        "title": "Name Data",
+        "required": ["firstName"],
     }
     builder = pjs.ObjectBuilder(schema)
-    ns = builder.build_classes() # by defualt strict = False
+    ns = builder.build_classes()  # by defualt strict = False
     NameData = ns.NameData
     # no strict flag - so should pass even no firstName
     NameData(lastName="hello")
@@ -482,12 +462,7 @@ def test_boolean_in_child_object():
         "$schema": "http://json-schema.org/schema#",
         "$id": "test",
         "type": "object",
-        "properties": {
-            "data": {
-                "type": "object",
-                "additionalProperties": True
-            }
-        }
+        "properties": {"data": {"type": "object", "additionalProperties": True}},
     }
     builder = pjs.ObjectBuilder(schema)
     ns = builder.build_classes()
@@ -495,25 +470,25 @@ def test_boolean_in_child_object():
     ns.Test(data={"my_bool": True})
 
 
-
-@pytest.mark.parametrize('default', [
-    '{"type": "boolean", "default": false}',
-    '{"type": "string", "default": "Hello"}',
-    '{"type": "integer", "default": 500}'
-])
+@pytest.mark.parametrize(
+    "default",
+    [
+        '{"type": "boolean", "default": false}',
+        '{"type": "string", "default": "Hello"}',
+        '{"type": "integer", "default": 500}',
+    ],
+)
 def test_default_values(default):
     default = json.loads(default)
     schema = {
         "$schema": "http://json-schema.org/schema#",
         "$id": "test",
         "type": "object",
-        "properties": {
-            "sample": default
-        }
+        "properties": {"sample": default},
     }
 
     builder = pjs.ObjectBuilder(schema)
     ns = builder.build_classes()
 
     x = ns.Test()
-    assert x.sample == default['default']
+    assert x.sample == default["default"]
